from forestadmin.datasource_toolkit.exceptions import DatasourceToolkitException


class DjangoDatasourceException(DatasourceToolkitException):
    pass


class DjangoNativeDriver(DjangoDatasourceException):
    pass



import importlib
import os

from django.conf import settings
from django.db import DEFAULT_DB_ALIAS, models
from django.db.backends.utils import CursorDebugWrapper
from forestadmin.datasource_django.exception import DjangoNativeDriver


class NativeDriverWrapper:
    def __init__(self, connection):
        self.connection = connection

        self._old_dj_allow_async_unsafe = None
        self._cursor = None

    def __enter__(self) -> CursorDebugWrapper:
        self._old_dj_allow_async_unsafe = os.environ.get("DJANGO_ALLOW_ASYNC_UNSAFE")
        os.environ["DJANGO_ALLOW_ASYNC_UNSAFE"] = "true"
        self._cursor = self.connection.cursor()
        return self._cursor

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._old_dj_allow_async_unsafe is None:
            del os.environ["DJANGO_ALLOW_ASYNC_UNSAFE"]
        else:
            os.environ["DJANGO_ALLOW_ASYNC_UNSAFE"] = self._old_dj_allow_async_unsafe

        self._cursor.close()


def get_db_for_native_driver(model: models.Model):
    db_name = DEFAULT_DB_ALIAS

    # if we use multiple databases with db router(s)
    if len(settings.DATABASES.keys()) > 1 and len(settings.DATABASE_ROUTERS) > 0:
        db_read = None
        db_write = None
        for router_str in settings.DATABASE_ROUTERS:
            module_name, class_name = router_str.rsplit(".", 1)
            router = getattr(importlib.import_module(module_name), class_name)()

            # forest_native_driver is an hint for user
            if db_read is None:
                db_read = router.db_for_read(model, forest_native_driver=True)
            if db_write is None:
                db_write = router.db_for_write(model, forest_native_driver=True)

            if db_read is not None and db_write is not None:
                break

        # because if it's not the same, we cannot choose which one to use
        if db_read != db_write:
            raise DjangoNativeDriver(
                "Cannot choose database between return db router. "
                f"Read database is '{db_read}', and write database is '{db_write}'."
            )
        db_name = db_read or DEFAULT_DB_ALIAS
    return db_name



from datetime import date, datetime
from decimal import Decimal
from typing import Any, cast

from django.db.models import Model
from forestadmin.datasource_django.interface import BaseDjangoCollection
from forestadmin.datasource_django.utils.polymorphic_util import DjangoPolymorphismUtil
from forestadmin.datasource_toolkit.interfaces.fields import (
    Column,
    PrimitiveType,
    is_polymorphic_many_to_one,
    is_polymorphic_one_to_one,
)
from forestadmin.datasource_toolkit.interfaces.query.projections import Projection
from forestadmin.datasource_toolkit.interfaces.query.projections.factory import ProjectionFactory
from forestadmin.datasource_toolkit.interfaces.records import RecordsDataAlias


def instance_to_record_data(
    instance: Model, projection: Projection, collection: BaseDjangoCollection
) -> RecordsDataAlias:
    record_data = {}
    for field_name in projection.columns:
        if DjangoPolymorphismUtil.is_type_field_of_generic_fk(field_name, collection):
            record_data[field_name] = DjangoPolymorphismUtil.get_collection_name_from_content_type(
                getattr(instance, field_name)
            )
        else:
            record_data[field_name] = serialize_value(
                getattr(instance, field_name), cast(Column, collection.schema["fields"][field_name])
            )

    for relation_name, subfields in projection.relations.items():
        relation = getattr(instance, relation_name, None)
        if relation is None:
            record_data[relation_name] = None
            continue

        relation_schema = collection.schema["fields"][relation_name]
        _projection = subfields
        foreign_collection_name = relation_schema.get("foreign_collection")

        if is_polymorphic_one_to_one(relation_schema):
            relation = relation.all()  #  type:ignore
            # when using relation.first, django make a request for ordering, so we avoid this extra request by using all
            if not relation.exists():
                record_data[relation_name] = None
                continue
            relation = relation[0]

        elif is_polymorphic_many_to_one(relation_schema):
            target_type = getattr(instance, relation_schema["foreign_key_type_field"], None)

            target_type = DjangoPolymorphismUtil.get_collection_name_from_content_type(target_type)
            foreign_collection = collection.datasource.get_collection(target_type)
            _projection = ProjectionFactory.all(foreign_collection, allow_nested=False)
            foreign_collection_name = foreign_collection.name

        foreign_collection = collection.datasource.get_collection(foreign_collection_name)  # type:ignore
        record_data[relation_name] = instance_to_record_data(
            relation,
            _projection,
            foreign_collection,
        )
    return record_data


def serialize_value(value: Any, column_schema: Column):
    if value is None:
        return value
    if column_schema["column_type"] == PrimitiveType.NUMBER:
        if isinstance(value, Decimal):
            return float(value)
        else:
            return value

    # Do we want to return date and datetime as object or isoformat ??
    elif column_schema["column_type"] == PrimitiveType.DATE_ONLY:
        return value if isinstance(value, date) else date.fromisoformat(value)
    elif column_schema["column_type"] == PrimitiveType.DATE:
        return value if isinstance(value, datetime) else datetime.fromisoformat(value)

    return value
